/* $Id: crires_wlcalib.c,v 1.49 2011/02/09 10:16:18 yjung Exp $
 *
 * This file is part of the CRIRES Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2011/02/09 10:16:18 $
 * $Revision: 1.49 $
 * $Name: crire-2_1_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_wlxcorr.h"
#include "irplib_ppm.h"
#include "irplib_plot.h"

#include "crires_wlcalib.h"
#include "crires_wlestimate.h"
#include "crires_combine.h"
#include "crires_extract.h"
#include "crires_dfs.h"
#include "crires_utils.h"
#include "crires_pfits.h"
#include "crires_load.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup crires_wlcalib     Wavelength calibration routines
 *
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration with a lamp
  @param    lamp        the lamp image
  @param    period
  @param    cat         the catalog
  @param    flat        the flat or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bpm or NULL
  @param    detlin      the detlin coefficients or NULL
  @param    wstart      Start wavelength at 1st pixel (or -1 if unknown)
  @param    wstop       Stop wavelength at last pixel (or -1 if unknown)
  @param    wmin        Minimum wavel. to consider (<0 if none)
  @param    wmax        Maximum wavel. to consider (<0 if none)
  @param    log_flag    flag to use the logarithm of the catalog
  @param    chip        the chip number
  @param    start_y     the lower part of the zone to extract
  @param    stop_y      the upper part of the zone to extract
  @param    degree      the requested degree
  @param    slitw       the slit width
  @param    fwhm        the fwhm
  @param    error       the wavelength error (in pixels)
  @param    samples     the number of samples
  @param    clean_spec      Flag to remove the low frequency part from the spec.
  @param    xc_limit        Lower value allowed for the XC factor
  @param    ppm         flag to use PPM
  @param    display     flag for display mode 
  @param    xcorr       the cross correlation factor (computed)
  @param    lines_flux  the lamp lines flux
  @param    fwhm_med    the lamp lines fwhm 
  @param    rpower      the lamp lines resolving power
  @return   the result polynomial or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlcalib_lamp(
        const char          *   lamp,
        crires_illum_period     period,
        const char          *   cat,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        double                  wstart,
        double                  wstop,
        double                  wmin,
        double                  wmax,
        int                     log_flag,
        int                     chip,
        int                     start_y,
        int                     stop_y,
        int                     degree,
        double                  slitw,
        double                  fwhm,
        double                  error,
        int                     samples,
        int                     clean_spec,
        double                  xc_limit,
        int                     ppm,
        int                     display,
        double              *   xcorr,
        double              *   lines_flux,
        double              *   fwhm_med,
        double              *   rpower)
{
    cpl_polynomial      *   wl_est ;
    cpl_table           *   catal ;
    cpl_bivector        *   catal_biv ;
    double              *   catal_biv_x ;
    double              *   catal_biv_y ;
    double                  val, dit ;
    cpl_propertylist    *   plist ;
    cpl_imagelist       *   lamplist ;
    cpl_imagelist       *   lamp_ima ;
    cpl_image           *   tmp_ima ;
    cpl_vector          *   extr ;
    int                     nrows ;
    cpl_polynomial      *   wl_sol ;
    double                  xc ;
    int                     ly ;
    int                     i ;
    
    /* Test entries */
    if (lamp == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    if (lines_flux == NULL || fwhm_med == NULL || rpower == NULL) return NULL ;

    /* Initialise */
    if (xcorr != NULL) *xcorr = -1 ;
    ly = -1 ;
    
    /* Check the central wavelength */
    if (wstart > 0 && wstop > 0) 
        wl_est = crires_wlestimate_compute(wstart, wstop) ;
    else 
        wl_est = crires_wlestimate_get(lamp, chip) ;

    /* No catalog, use the estimate */
    if (cat == NULL) {
        cpl_msg_error(__func__, "The catalog is missing") ;
        return wl_est ;
    }
    
    /* Load the  catalog */
    catal = crires_load_table_check(cat, 1, CRIRES_PROTYPE_CATALOG, -1, -1, 0) ;
    if (catal == NULL) {
        cpl_msg_info(__func__, "The catalog cannot be loaded") ;
        return wl_est ;
    }
    
    /* Create the catalog spectrum in a bivector */
    nrows = cpl_table_get_nrow(catal) ;
    catal_biv = cpl_bivector_new(nrows) ;
    catal_biv_x = cpl_bivector_get_x_data(catal_biv) ;
    catal_biv_y = cpl_bivector_get_y_data(catal_biv) ;
    for (i=0 ; i<nrows ; i++) {
        catal_biv_x[i] = cpl_table_get(catal, CRIRES_COL_WAVELENGTH, i, NULL) ;
        val = cpl_table_get(catal, CRIRES_COL_EMISSION, i, NULL) ;
        if (log_flag && val > 0)    catal_biv_y[i] = log10(val) ;
        else                        catal_biv_y[i] = val ;
    }
    cpl_table_delete(catal) ;

    /* Load the lamp */
    if ((lamplist = crires_load_file(lamp, period, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the lamp") ;
        cpl_bivector_delete(catal_biv) ;
        return wl_est ;
    }

    /* Get the DIT value from the header */
    if ((plist=cpl_propertylist_load(lamp, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from frame");
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(lamplist) ;
        return wl_est ;
    }
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot get the DIT from the header") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(lamplist) ;
        return wl_est ;
    }
    
    /* Apply the calibration */
    lamp_ima = cpl_imagelist_new() ;
    tmp_ima = cpl_image_duplicate(cpl_imagelist_get(lamplist, chip-1)) ;
    cpl_imagelist_set(lamp_ima, tmp_ima, 0) ;
    crires_calib_chip_list(lamp_ima, period, chip, flat, dark, bpm, detlin,dit);
    cpl_imagelist_delete(lamplist) ;

   /* Get the correction for the position */
    if (chip == 1) ly = crires_get_detector_ly1(period) ;
    if (chip == 2) ly = crires_get_detector_ly2(period) ;
    if (chip == 3) ly = crires_get_detector_ly3(period) ;
    if (chip == 4) ly = crires_get_detector_ly4(period) ;

    /* Extract the spectrum to calibrate */
    if ((extr=crires_extract_lines(cpl_imagelist_get(lamp_ima, 0),
                    start_y-ly, stop_y-ly, 0, 0)) == NULL) {
        cpl_msg_error(__func__, "Cannot extract the lamp") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(lamp_ima) ;
        return wl_est ;
    }
    cpl_imagelist_delete(lamp_ima) ;

    /* Calibration */
    if ((wl_sol = crires_wlcalib_engine(extr, catal_biv, wl_est, slitw,
                    fwhm, degree, wmin, wmax, error, samples, clean_spec, 
                    xc_limit, ppm, display, &xc, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the wavelength calibration") ;
        cpl_vector_delete(extr) ;
        cpl_bivector_delete(catal_biv) ;
        return wl_est ;
    }
    if (xcorr != NULL) *xcorr = xc ;
    cpl_bivector_delete(catal_biv) ;
 
    /* Compute the QC for the lamp lines */
    cpl_msg_info(__func__, "Compute the Lamp qc parameters") ;
    if (crires_extract_lines_properties(extr, wl_sol, display, 
                lines_flux, fwhm_med, rpower) == -1) {
        cpl_msg_warning(__func__, "Cannot compute the THAR qc parameters") ;
    }
    cpl_vector_delete(extr) ;

    /* Plot if requested */
    if (display) {
        irplib_wlxcorr_plot_solution(wl_est, wl_sol, NULL, 1, 1024);
    }
    cpl_polynomial_delete(wl_est) ;
    
    return wl_sol ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration with a sky
  @param    sky         the sky image
  @param    period
  @param    oh          the oh catalog
  @param    flat        the flat or NULL
  @param    dark        the dark or NULL
  @param    bpm         the bpm or NULL
  @param    detlin      the detlin coefficients or NULL
  @param    wstart      Start wavelength at 1st pixel (or -1 if unknown)
  @param    wstop       Stop wavelength at last pixel (or -1 if unknown)
  @param    wmin        Minimum wavel. to consider (<0 if none)
  @param    wmax        Maximum wavel. to consider (<0 if none)
  @param    log_flag    flag to use the logarithm of the oh catalog
  @param    hitran      the hitran catalog
  @param    chip        the chip number
  @param    start_y     the lower part of the zone to extract
  @param    stop_y      the upper part of the zone to extract
  @param    degree      the requested degree
  @param    slitw       the slit width
  @param    fwhm        the fwhm
  @param    error       the wavelength error (in pixels)
  @param    samples     the number of samples
  @param    clean_spec      Flag to remove the low frequency part from the spec.
  @param    xc_limit        Lower value allowed for the XC factor
  @param    ppm         flag to use PPM
  @param    display     flag for display mode 
  @param    xcorr       the cross correlation factor (computed)
  @return   the result polynomial or NULL in error case
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlcalib_sky(
        const char          *   sky,
        crires_illum_period     period,
        const char          *   oh,
        const char          *   flat,
        const char          *   dark,
        const char          *   bpm,
        const char          *   detlin,
        double                  wstart,
        double                  wstop,
        double                  wmin,
        double                  wmax,
        int                     log_flag,
        const char          *   hitran,
        int                     chip,
        int                     start_y,
        int                     stop_y,
        int                     degree,
        double                  slitw,
        double                  fwhm,
        double                  error,
        int                     samples,
        int                     clean_spec,
        double                  xc_limit,
        int                     ppm,
        int                     display,
        double              *   xcorr)
{
    double                  wlen ;
    cpl_polynomial      *   wl_est ;
    cpl_table           *   catal ;
    cpl_bivector        *   catal_biv ;
    double              *   catal_biv_x ;
    double              *   catal_biv_y ;
    double                  val, dit ;
    cpl_imagelist       *   skylist ;
    cpl_imagelist       *   sky_ima ;
    cpl_image           *   tmp_ima ;
    cpl_propertylist    *   plist ;
    cpl_vector          *   extr ;
    int                     nrows, ly ;
    cpl_polynomial      *   wl_sol ;
    double                  xc ;
    int                     i ;
    
    /* Test entries */
    if (sky == NULL) return NULL ;
    if (period == CRIRES_ILLUM_UNKNOWN) return NULL ;
    
    /* Initialise */
    if (xcorr != NULL) *xcorr = -1 ;
    ly = -1 ;
    
    /* Check the central wavelength */
    if (wstart > 0 && wstop > 0) 
        wl_est = crires_wlestimate_compute(wstart, wstop) ;
    else 
        wl_est = crires_wlestimate_get(sky, chip) ;
    wlen = cpl_polynomial_eval_1d(wl_est, 512.0, NULL) ;

    /* No catalog, use the estimate */
    if (oh == NULL && hitran == NULL) {
        cpl_msg_error(__func__, "The catalogs are missing") ;
        return wl_est ;
    }
    
    /* Load the relevant catalog */
    if (wlen < 1800) {
        if (oh == NULL) {
            cpl_msg_error(__func__, "Need the OH catalog below 1.8 microns") ;
            return wl_est ;
        }
        catal = crires_load_table_check(oh, 1, CRIRES_PROTYPE_CATALOG, -1,-1,0);
    } else {
        if (hitran == NULL) {
            cpl_msg_error(__func__,"Need the HITRAN catalog above 1.8 microns");
            return wl_est ;
        }
        catal = crires_load_table_check(hitran, 1, CRIRES_PROTYPE_CATALOG, 
                -1, -1, 0) ;
    }
    
    /* Create the catalog spectrum in a bivector */
    nrows = cpl_table_get_nrow(catal) ;
    catal_biv = cpl_bivector_new(nrows) ;
    catal_biv_x = cpl_bivector_get_x_data(catal_biv) ;
    catal_biv_y = cpl_bivector_get_y_data(catal_biv) ;
    for (i=0 ; i<nrows ; i++) {
        catal_biv_x[i] = cpl_table_get(catal, CRIRES_COL_WAVELENGTH, i, NULL) ;
        val = cpl_table_get(catal, CRIRES_COL_EMISSION, i, NULL) ;
        if (log_flag && val > 0)    catal_biv_y[i] = log10(val) ;
        else                        catal_biv_y[i] = val ;
    }
    cpl_table_delete(catal) ;

    /* Load the sky */
    if ((skylist = crires_load_file(sky, period, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load the sky") ;
        cpl_bivector_delete(catal_biv) ;
        return wl_est ;
    }

    /* Get the DIT value from the header */
    if ((plist=cpl_propertylist_load(sky, 0)) == NULL) {
        cpl_msg_error(__func__, "getting header from frame");
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(skylist) ;
        return wl_est ;
    }
    dit = crires_pfits_get_dit(plist) ;
    cpl_propertylist_delete(plist) ;
    if (cpl_error_get_code() != CPL_ERROR_NONE) {
        cpl_msg_error(__func__, "Cannot get the DIT from the header") ;
        cpl_bivector_delete(catal_biv) ;
        cpl_imagelist_delete(skylist) ;
        return wl_est ;
    }

    /* Apply the calibration */
    sky_ima = cpl_imagelist_new() ;
    tmp_ima = cpl_image_duplicate(cpl_imagelist_get(skylist, chip-1)) ;
    cpl_imagelist_set(sky_ima, tmp_ima, 0) ;
    crires_calib_chip_list(sky_ima, period, i+1, flat, dark, bpm, detlin, dit) ;
    cpl_imagelist_delete(skylist) ;

    /* Extract the spectrum to calibrate */
    if (start_y > 0 && stop_y  > 0) {

        /* Get the correction for the position */
        if (chip == 1) ly = crires_get_detector_ly1(period) ;
        if (chip == 2) ly = crires_get_detector_ly2(period) ;
        if (chip == 3) ly = crires_get_detector_ly3(period) ;
        if (chip == 4) ly = crires_get_detector_ly4(period) ;

        /* Extract where it was specified */
        if ((extr=crires_extract_lines(cpl_imagelist_get(sky_ima, 0),
                        start_y-ly, stop_y-ly, 0, 0)) == NULL) {
            cpl_msg_error(__func__, "Cannot extract the sky") ;
            cpl_bivector_delete(catal_biv) ;
            cpl_imagelist_delete(sky_ima) ;
            return wl_est ;
        }
    } else {
        /* Intelligent extraction (with bright line removal) */
        if ((extr=crires_extract_lines_spectrum(cpl_imagelist_get(sky_ima, 0), 
                        period, chip, 9, 20, 0, 0, -1, -1)) == NULL) {
            cpl_msg_error(__func__, "Cannot extract the sky") ;
            cpl_bivector_delete(catal_biv) ;
            cpl_imagelist_delete(sky_ima) ;
            return wl_est ;
        }
    }
    cpl_imagelist_delete(sky_ima) ;

    /* Calibration */
    if ((wl_sol = crires_wlcalib_engine(extr, catal_biv, wl_est, slitw,
                    fwhm, degree, wmin, wmax, error, samples, clean_spec, 
                    xc_limit, ppm, display, &xc, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot compute the wavelength calibration") ;
        cpl_vector_delete(extr) ;
        cpl_bivector_delete(catal_biv) ;
        return wl_est ;
    }
    if (xcorr != NULL) *xcorr = xc ;
    cpl_vector_delete(extr) ;
    cpl_bivector_delete(catal_biv) ;
    
    /* Plot if requested */
    if (display) {
        irplib_wlxcorr_plot_solution(wl_est, wl_sol, NULL, 1, 1024);
    }
    cpl_polynomial_delete(wl_est) ;
    
    return wl_sol ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration using a spectrum
  @param    spec            the spectrum to calibrate
  @param    catalog_spec    the spectrum obtained with the catalog lines
  @param    phdisp          the physical model polynomial
  @param    slit_width      Width in pixels of the slit used
  @param    fwhm
  @param    degree          Degree of the fitted polynomial
  @param    wl_min          The Minimum wavelength to consider
  @param    wl_max          The Maximum wavelength to consider
  @param    wl_err          The wl error (in pixels)
  @param    nsamples        Number of samples for wavelength computation
  @param    clean_spec      Flag to remove the low frequency part from the spec.
  @param    xc_limit        Lower value allowed for the XC factor
  @param    use_ppm         Use the point pattern matching at the end
  @param    display         flag to display intermediate results
  @param    xcorr           the cross correlation factor (computed)
  @return   the wavelength polynomial or NULL in error case

  The returned polynomial has to be deallocated with cpl_polynomial_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlcalib_engine(
        const cpl_vector        *   spec,
        const cpl_bivector      *   catalog_spec,
        const cpl_polynomial    *   phdisp,
        double                      slit_width,
        double                      fwhm,
        int                         degree,
        double                      wl_min,
        double                      wl_max,
        double                      wl_err,
        int                         nsamples,
        int                         clean_spec,
        double                      xc_limit,
        int                         use_ppm,
        int                         display,
        double                  *   xcorr,
        cpl_table               **  wlres)
{
    double                  wl_err_wl ;
    double                  wl_start, wl_stop ;
    int                     pix_start, pix_stop, pix_min, pix_max ;
    cpl_vector          *   spec_extract ;
    cpl_vector          *   spec_clean ;
    double              *   pspec ;
    int                     filt_size ;
    double                  xc ;
    cpl_vector          *   wl_errors ;
    cpl_polynomial      *   xcdisp ;
    cpl_polynomial      *   xcdisp_tmp ;
    cpl_polynomial      *   wl_poly_ppm ;
    cpl_vector          *   filtered ;
    double                  threshold, median, stdev ;
    cpl_bivector        *   catal_loc ;
    double              *   catal_loc_wl ;
    double              *   catal_loc_intens ;
    int                     nsam_factor, wlerr_factor ;
    cpl_table           *   spc_table ;
    cpl_vector          *   xcorrs ;
    cpl_table           *   wlres_loc ;
    int                     i ;

    /* Test entries */
    if (spec == NULL || catalog_spec == NULL || phdisp == NULL) return NULL ;
    if (degree < 1 || degree > 4) return NULL ;
    if (nsamples < 10)  nsam_factor = 1 ;
    else                nsam_factor = 10 ;
    wlerr_factor = 5 ;

    /* Initialise */
    filt_size = 9 ;
    if (wlres != NULL) *wlres = NULL ;

    /* Convert the wavelength error from pixels to wavelength */
    wl_err_wl = cpl_polynomial_eval_1d(phdisp, wl_err, NULL) ;
    wl_err_wl -= cpl_polynomial_eval_1d(phdisp, 0.0, NULL) ;
    cpl_msg_info(__func__, "Wavelength error : %g pix = %g nm", wl_err,
            wl_err_wl) ;

    /* Define boundaries */
    pix_start = 1 ;
    pix_stop = cpl_vector_get_size(spec) ;
    wl_start = cpl_polynomial_eval_1d(phdisp, (double)pix_start, NULL) ;
    wl_stop = cpl_polynomial_eval_1d(phdisp, (double)pix_stop, NULL) ;
    if ((wl_min > 0) && (wl_min > wl_start)) {
        pix_min = pix_stop - 
            (int)((pix_stop-pix_start)*(wl_stop-wl_min)/(wl_stop-wl_start)) ;
        cpl_msg_warning(__func__, "Start on pixel %d\n", pix_min) ;
    } else pix_min = pix_start ;
    if ((wl_max > 0) && (wl_max < wl_stop)) {
        pix_max = pix_start + 
            (int)((pix_stop-pix_start)*(wl_max-wl_start)/(wl_stop-wl_start)) ;
        cpl_msg_warning(__func__, "Stop on pixel %d\n", pix_max) ;
    } else pix_max = pix_stop ;
    /* Check */
    if (pix_max <= pix_min) {
        cpl_msg_error(__func__, "Wrong boundaries %d pix -> %d pix", 
                pix_min, pix_max) ;
        return NULL ;
    }

    /* Zero the rejected parts */
    spec_extract = cpl_vector_duplicate(spec) ;
    pspec = cpl_vector_get_data(spec_extract) ;
    for (i=0 ; i<cpl_vector_get_size(spec) ; i++) {
        if (i+1 < pix_min) pspec[i] = 0.0 ;
        if (i+1 > pix_max) pspec[i] = 0.0 ;
    }

    /* Clean the spectrum from the low frequency signal if requested */
    if (clean_spec) {
        cpl_msg_info(__func__, "Low Frequency signal removal") ;
        cpl_msg_indent_more() ;
        /* Subrtract the low frequency part */
        if ((filtered=cpl_vector_filter_median_create(spec_extract, 
                        filt_size))==NULL){
            cpl_msg_error(__func__, "Cannot filter the spectrum") ;
            cpl_msg_indent_less() ;
            return NULL ;
        }
        spec_clean = cpl_vector_duplicate(spec_extract) ;
        cpl_vector_subtract(spec_clean, filtered) ;
        cpl_vector_delete(filtered) ;
        cpl_msg_indent_less() ;
    } else {
        spec_clean = cpl_vector_duplicate(spec_extract) ;
    }
    cpl_vector_delete(spec_extract) ;
    
    /* Extract the interesting part of the catalog */
    cpl_msg_info(__func__, "Sub-Catalog extraction") ;
    cpl_msg_indent_more() ;
    catal_loc = irplib_wlxcorr_cat_extract(catalog_spec,
            wl_start-wl_err_wl, wl_stop+wl_err_wl) ;
    if (catal_loc == NULL) {
        cpl_msg_error(__func__, "Cannot extract sub-catalog [%g %g]",
                wl_start, wl_stop) ;
        /* Should not raise an error if there is no line */
        cpl_error_reset() ;
        /********/
        cpl_msg_indent_less() ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_indent_less() ;
  
    /* Zero the rejected parts */
    catal_loc_wl = cpl_bivector_get_x_data(catal_loc) ;
    catal_loc_intens = cpl_bivector_get_y_data(catal_loc) ;
    for (i=0 ; i<cpl_bivector_get_size(catal_loc) ; i++) {
        if (wl_min > 0) {
            if (catal_loc_wl[i] < wl_min) catal_loc_intens[i] = 0.0 ;
        }
        if (wl_max > 0) {
            if (catal_loc_wl[i] > wl_max) catal_loc_intens[i] = 0.0 ;
        }
    }

    /* Plot inputs */
    if (display) {
        irplib_wlxcorr_catalog_plot(catal_loc, wl_start, wl_stop) ;
        irplib_vector_plot(
    "set grid;set xlabel 'Position (Pixel)';set ylabel 'Intensity (ADU/sec)';",
    "t 'Extracted spectrum' w lines", "", spec_clean);
    }

    /* First solution */
    cpl_msg_info(__func__, "Find the best linear solution") ;
    cpl_msg_indent_more() ;
    wl_errors = cpl_vector_new(2) ;
    cpl_vector_fill(wl_errors, wl_err_wl) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", pow(nsamples, 2)) ;
    if ((xcdisp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, 1, phdisp, 
                    wl_errors, nsamples, slit_width, fwhm, &xc, 
                    NULL, &xcorrs)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate") ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        if (xcorrs != NULL) cpl_vector_delete(xcorrs) ;
        cpl_error_reset() ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
   
    /* Plot the first correlation values */
    if (display) {
        irplib_vector_plot("set grid;", 
                "t 'Correlation values for the linear solution search' w lines",
                "", xcorrs) ;
    }
    if (xcorrs != NULL) cpl_vector_delete(xcorrs) ;
    cpl_vector_delete(wl_errors) ;

    /* Higher degree --> less samples  + Lower error --> best precision */
    cpl_msg_info(__func__, "Find the best degree-%d polynomial", degree) ;
    cpl_msg_indent_more() ;
    wl_errors = cpl_vector_new(degree+1) ;
    cpl_vector_fill(wl_errors, wl_err_wl/wlerr_factor) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err/wlerr_factor) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples/nsam_factor) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", 
            pow(nsamples/nsam_factor, degree+1)) ;
    if ((xcdisp_tmp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, degree, 
                    xcdisp, wl_errors, nsamples/nsam_factor, slit_width, fwhm,
                    &xc, NULL, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate with the %d degree", degree) ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_polynomial_delete(xcdisp) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
    cpl_polynomial_delete(xcdisp) ;
    xcdisp = xcdisp_tmp ;

    /* Lower error --> best precision */
    cpl_msg_info(__func__, "Refine the degree-%d polynomial", degree) ;
    cpl_msg_indent_more() ;
    cpl_vector_divide_scalar(wl_errors, wlerr_factor) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err/pow(wlerr_factor,2)) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples/nsam_factor) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", 
            pow(nsamples/nsam_factor, degree+1)) ;
    if ((xcdisp_tmp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, degree, 
                    xcdisp, wl_errors, nsamples/nsam_factor, slit_width, fwhm,
                    &xc, NULL, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate with the %d degree", degree) ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_polynomial_delete(xcdisp) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
    cpl_polynomial_delete(xcdisp) ;
    xcdisp = xcdisp_tmp ;

    /* Lower error --> best precision */
    cpl_msg_info(__func__, "Refine the degree-%d polynomial", degree) ;
    cpl_msg_indent_more() ;
    cpl_vector_divide_scalar(wl_errors, wlerr_factor) ;
    cpl_msg_info(__func__, "    Wl error: %g nm (%g pix)",
            cpl_vector_get_mean(wl_errors), wl_err/pow(wlerr_factor,3)) ;
    cpl_msg_info(__func__, "    Nb of samples: %d", nsamples/nsam_factor) ;
    cpl_msg_info(__func__, "    Nb of candidates: %g", 
            pow(nsamples/nsam_factor, degree+1)) ;
    if ((xcdisp_tmp=irplib_wlxcorr_best_poly(spec_clean, catal_loc, degree, 
                    xcdisp, wl_errors, nsamples/nsam_factor, slit_width, fwhm,
                    &xc, &wlres_loc, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot calibrate with the %d degree", degree) ;
        cpl_msg_indent_less() ;
        cpl_vector_delete(wl_errors) ;
        cpl_polynomial_delete(xcdisp) ;
        cpl_bivector_delete(catal_loc) ;
        cpl_vector_delete(spec_clean) ;
        return NULL ;
    }
    cpl_msg_info(__func__, "Cross-Correlation factor: %g", xc) ;
    cpl_msg_indent_less() ;
    cpl_polynomial_delete(xcdisp) ;
    xcdisp = xcdisp_tmp ;
    cpl_vector_delete(wl_errors) ;

    /* Plot results table */
    if (display) {
        /* Create the spc_table  */
        if ((spc_table = irplib_wlxcorr_gen_spc_table(spec_clean, catal_loc,
                        slit_width, fwhm, phdisp, xcdisp)) == NULL) {
            cpl_msg_error(cpl_func, "Cannot generate infos table") ;
            cpl_polynomial_delete(xcdisp) ;
            cpl_bivector_delete(catal_loc) ;
            cpl_vector_delete(spec_clean) ;
            cpl_table_delete(wlres_loc) ;
            return NULL ;
        }
        /* Plot */
        irplib_wlxcorr_plot_spc_table(spc_table, "XC") ;
        cpl_table_delete(spc_table) ;
    }
    cpl_bivector_delete(catal_loc) ;
    
    /* Test Cross correlation factor */
    if (xc < xc_limit) {
        cpl_msg_warning(__func__, 
                "Cross-correlation too low (%g). Use estimate", xc) ;
        *xcorr = -1.0 ;
        cpl_polynomial_delete(xcdisp) ;
        xcdisp = cpl_polynomial_duplicate(phdisp) ;
    } else {
        cpl_msg_info(__func__, "Cross-correlation: %g", xc) ;
        *xcorr = xc ;
    }

    /* If PPM has been requested */
    if (use_ppm) {
        /* Threshold the clean spectrum */
        cpl_msg_info(__func__, "Threshold the spectrum") ;
        stdev = cpl_vector_get_stdev(spec_clean) ;
        median = cpl_vector_get_median_const(spec_clean) ;
        threshold = median + 0.2 * stdev ;
        pspec = cpl_vector_get_data(spec_clean) ;
        for (i=0 ; i<cpl_vector_get_size(spec_clean) ; i++) {
            if (pspec[i] < threshold) pspec[i] = 0.0 ;
        }
 
        /* Wavelength calibration with the Point-Pattern matching */
        cpl_msg_info(__func__, "Point Pattern Matching") ;
        if ((wl_poly_ppm = irplib_ppm_engine(spec_clean, catalog_spec, xcdisp,
                        5, 5, 0.2, degree, display, NULL)) == NULL) {
            cpl_msg_warning(__func__, "Cannot calibrate") ;
            wl_poly_ppm = xcdisp ;
        } else cpl_polynomial_delete(xcdisp) ;
    } else {
        wl_poly_ppm = xcdisp ;
    }
    cpl_vector_delete(spec_clean) ;

    /* Return the wlres table if requested */
    if (wlres != NULL) *wlres = wlres_loc ;
    else cpl_table_delete(wlres_loc) ;

    return wl_poly_ppm ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration table generation
  @param    wl_poly     The wavelength solutions for the 4 chips
  @return   the wavelength tables or NULL in error case

  The returned table has to be deallocated with cpl_table_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_table ** crires_wlcalib_gen_wltab(const cpl_polynomial ** wl_poly)
{
    int             p[5] ;
    double          a, b, c, d, e ;
    cpl_table   **  wl_tab ;
    int             i ;
    
    /* Test entries */
    if (wl_poly == NULL) return NULL ;

    /* Initialise */
    p[0] = 0 ; p[1] = 1 ; p[2] = 2 ; p[3] = 3 ; p[4] = 4 ;
    wl_tab = cpl_malloc(CRIRES_NB_DETECTORS * sizeof(cpl_table*)) ;
    /* Loop on the chips */
    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        wl_tab[i] = cpl_table_new(1) ;
        cpl_table_new_column(wl_tab[i], CRIRES_COL_WL_COEFFA, CPL_TYPE_DOUBLE) ;
        cpl_table_new_column(wl_tab[i], CRIRES_COL_WL_COEFFB, CPL_TYPE_DOUBLE) ;
        cpl_table_new_column(wl_tab[i], CRIRES_COL_WL_COEFFC, CPL_TYPE_DOUBLE) ;
        cpl_table_new_column(wl_tab[i], CRIRES_COL_WL_COEFFD, CPL_TYPE_DOUBLE) ;
        cpl_table_new_column(wl_tab[i], CRIRES_COL_WL_COEFFE, CPL_TYPE_DOUBLE) ;
        if (wl_poly[i] != 0) {
            /* Construct the polynomial */
            a = cpl_polynomial_get_coeff(wl_poly[i], &(p[0])) ;
            b = cpl_polynomial_get_coeff(wl_poly[i], &(p[1])) ;
            c = cpl_polynomial_get_coeff(wl_poly[i], &(p[2])) ;
            d = cpl_polynomial_get_coeff(wl_poly[i], &(p[3])) ;
            e = cpl_polynomial_get_coeff(wl_poly[i], &(p[4])) ;
        } else {
            a = b = c = d = e = 0.0 ;
        }
        /* Store the result */
        cpl_table_set_double(wl_tab[i], CRIRES_COL_WL_COEFFA, 0, a) ;
        cpl_table_set_double(wl_tab[i], CRIRES_COL_WL_COEFFB, 0, b) ;
        cpl_table_set_double(wl_tab[i], CRIRES_COL_WL_COEFFC, 0, c) ;
        cpl_table_set_double(wl_tab[i], CRIRES_COL_WL_COEFFD, 0, d) ;
        cpl_table_set_double(wl_tab[i], CRIRES_COL_WL_COEFFE, 0, e) ;
    }
    return wl_tab ; 
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration map generation
  @param    wl_tab      table with the wavelengths solutions
  @return   the wavelength map images or NULL in error case

  The returned image has to be deallocated with cpl_imagelist_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * crires_wlcalib_gen_wlmap(const cpl_table ** wl_tab)
{
    cpl_imagelist       *   wl_map_list ;
    cpl_image           *   wl_map ;
    float               *   pwl_map ;
    double                  val ;
    int                     nx, ny ;
    cpl_polynomial      *   wl_poly ;
    int                     i, j, k ;

    /* Test entries */
    if (wl_tab == NULL) return NULL ;

    /* Initialise */
    nx = 1024 ;
    ny = 512 ;

    /* Create the output image list */
    wl_map_list = cpl_imagelist_new() ;

    for (i=0 ; i<CRIRES_NB_DETECTORS ; i++) {
        /* Create the image */
        wl_map = cpl_image_new(nx, ny, CPL_TYPE_FLOAT) ;
        pwl_map = cpl_image_get_data_float(wl_map) ;

        /* Get the polynomial */
        wl_poly = crires_wlcalib_gen_wlpoly(wl_tab[i]) ;

        /* Fill the image */
        for (j=0 ; j<nx ; j++) {
            if (wl_poly == NULL) val = 0.0 ;
            else val = cpl_polynomial_eval_1d(wl_poly, (double)j+1, NULL) ;
            for (k=0 ; k<ny ; k++) pwl_map[j + k*nx] = val ;
        }
        if (wl_poly != NULL) cpl_polynomial_delete(wl_poly) ;

        /* Put the image in the list */
        cpl_imagelist_set(wl_map_list, wl_map, i) ;
    }
    return wl_map_list ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Wavelength calibration polynomial generation
  @param    wl_tab      table with the wavelengths solutions
  @return   the wavelength polynomial or NULL in error case

  The returned polynomial has to be deallocated with cpl_polynomial_delete().
 */
/*----------------------------------------------------------------------------*/
cpl_polynomial * crires_wlcalib_gen_wlpoly(const cpl_table * wl_tab)
{
    cpl_polynomial      *   wl_poly ;
    double                  a, b, c, d, e ;
    int                     nrows, power ;

    /* Test entries */
    if (wl_tab == NULL) return NULL ;

    /* Initialise */
    nrows = cpl_table_get_nrow(wl_tab) ;
    if (nrows != 1) return NULL ;

    /* Get the coefficients */
    a = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFA, 0, NULL) ;
    b = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFB, 0, NULL) ;
    c = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFC, 0, NULL) ;
    d = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFD, 0, NULL) ;
    e = cpl_table_get_double(wl_tab, CRIRES_COL_WL_COEFFE, 0, NULL) ;

    /* Create the polynomial */
    wl_poly = cpl_polynomial_new(1) ;
    power = 0 ; cpl_polynomial_set_coeff(wl_poly, &power, a) ;
    power = 1 ; cpl_polynomial_set_coeff(wl_poly, &power, b) ;
    power = 2 ; cpl_polynomial_set_coeff(wl_poly, &power, c) ;
    power = 3 ; cpl_polynomial_set_coeff(wl_poly, &power, d) ;
    power = 4 ; cpl_polynomial_set_coeff(wl_poly, &power, e) ;
    
    return wl_poly ;
}

/**@}*/
